/*/////////
//
//  varlib.c 
//
//      AUTHOR: Bruce's varlib.c modified by Russell Lowke
//   CSCIE-215: Assignment Five
//
//       Usage: a simple storage system to store name=value pairs
//              with facility to mark items as part of the environment
//
//   interface: VLstore( name, value )    returns 1 for 0k, 0 for no
//              VLlookup( name )          returns string or NULL if not there
//              VLset()                   prints out current table
//              VLexport( name )          adds name to list of env vars
//              VLtable2environ()         copy from table to environ
//              VLenviron2table()         copy from environ to table
//
//     details: the table is stored as an array of structs that
//	        contain a flag for `global' and a single string of
//	        the form name=value.  This allows EZ addition to the
//              environment.  It makes searching pretty easy, as
//              long as you search for "name=" 
//
*/

#include	"varlib.h"

#define	MAXVARS	200		/* a linked list would be nicer */

struct var
	{
	char *str;		/* name=val string	*/
	int  global;		/* a boolean		*/
	};

/* local vars: the table */
static struct var tab[MAXVARS];

/* local functions */
static char *new_string( char *, char *);
static struct var *find_item(char *, int);


/*
//  VLstore
//
//     traverse list, if found, replace it, else add at end
//     since there is no delete, a blank one is a free one
//     return 0 if trouble, 1 if ok
//
//     args:  *name         --  var name to store
//            *val          --  value to set
*/
int VLstore( char *name, char *val )
{
    struct var *itemp;
    char	*s;
        
    if ( (itemp = find_item(name,1)) == NULL )      /* find spot to put it    */
        return 0;

    if ( itemp->str )                               /* chuck old value        */
        free(itemp->str);

    if ( ( s = new_string( name, val )) == NULL )   /* new name=val           */
        return 0;				    /* no memory              */
    
    itemp->str = s;                                 /* store it               */
    return 1;
}


/*
//  new_string
//
//     returns new string of form name=value or NULL on error
//
//     args:  *name        --  name of variable
//            *val         --  value of variable
*/
char *new_string( char *name, char *val )
{
    char	*retval;

    retval = malloc( strlen(name) + strlen(val) + 2 );
    if ( retval != NULL )
        sprintf(retval, "%s=%s", name, val );
    return retval;
}


/*
//  VLlookup
//
//     returns value of var or empty string if not there
//
//     args:  *name         --  name of var to look up
*/
char *VLlookup( char *name )
{
    struct var *itemp;

    if ( (itemp = find_item(name,0)) != NULL )
        return itemp->str + 1 + strlen(name);
    return "";
}


/*
//  VLexport
//
//     marks a var for export, adds it if not there
//     returns 0 for no, 1 for ok
//
//     args:  *name     --  name of var to export
*/
int VLexport( char *name )
{
	struct var *itemp;

	if ( (itemp = find_item(name,0)) != NULL ){
		itemp->global = 1;
		return 1;
	}
	if ( VLstore(name, "") == 1 )
		VLexport(name);
	return 0;
}


/*
//  find_item
//
//     searches table for an item
//     returns ptr to struct or NULL if not found
//     OR if (first_blank) then ptr to first blank one
//
//     args:  *name         --  variable name to look up
//            first_blank   --  ptr to first blank one
*/
static struct var *find_item( char *name , int first_blank )
{
    int   i;
    int   len = strlen(name);
    char *s;

    for( i = 0 ; i<MAXVARS && tab[i].str != NULL ; i++ )
    {
        s = tab[i].str;
        if ( strncmp(s,name,len) == 0 && s[len] == '=' ){
            return &tab[i];
        }
    }
    if ( i < MAXVARS && first_blank )
        return &tab[i];
    return NULL;
}


/*
//  VLset
//
//     performs the shell's  `set'  command
//     Lists the contents of the variable table, marking each
//     exported variable with the symbol  '[E]' 
*/
void VLset()
{
	int	i;
	for(i = 0 ; i<MAXVARS && tab[i].str != NULL ; i++ )
	{
		if ( tab[i].global )
			printf("[E] %s\n", tab[i].str);
		else
			printf("    %s\n", tab[i].str);
	}
}


/*
//  VLenviron2table
//
//     initialize the variable table by loading array of strings
//     return 1 for ok, 0 for not ok
//
//     args:  *env[]         --  environment list
*/
int VLenviron2table(char *env[])
{
	int     i;
	char	*newstring;

	for(i = 0 ; env[i] != NULL ; i++ )
	{
		if ( i == MAXVARS )
			return 0;
		newstring = malloc(1+strlen(env[i]));
		if ( newstring == NULL )
			return 0;
		strcpy(newstring, env[i]);
		tab[i].str = newstring;
		tab[i].global = 1;
	}
	while( i < MAXVARS ){		/* I know we don't need this	*/
		tab[i].str = NULL ;	/* static globals are nulled	*/
		tab[i++].global = 0;	/* by default			*/
	}
	return 1;
}


/*
//  VLtable2environ
//
//     build an array of pointers suitable for making a new environment
//     note, you need to free() this when done to avoid memory leaks
*/
char **VLtable2environ()
{
	int	i,			/* index			*/
		j,			/* another index		*/
		n = 0;			/* counter			*/
	char	**envtab;		/* array of pointers		*/

	/*
	// first, count the number of global variables
	*/

	for( i = 0 ; i<MAXVARS && tab[i].str != NULL ; i++ )
		if ( tab[i].global == 1 )
			n++;

	/* then, allocate space for that many variables	*/
	envtab = (char **) malloc( (n+1) * sizeof(char *) );
	if ( envtab == NULL )
		return NULL;

	/* then, load the array with pointers		*/
	for(i = 0, j = 0 ; i<MAXVARS && tab[i].str != NULL ; i++ )
		if ( tab[i].global == 1 )
			envtab[j++] = tab[i].str;
	envtab[j] = NULL;
	return envtab;
}
